<?php
namespace Service;

use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\DBAL\Connection;
use Doctrine\ORM\Query\ParameterTypeInferer;
use Doctrine\ORM\EntityManager;
use Service\Manager\Mongo\Types\Type;
use Util\OpensslCrypt;
use Util\RijndaelCrypt;
use Util\Utilities;
use Zeedhi\Framework\DataSource\FilterCriteria;
use \Service\Manager\Mongo\Manager;
use Util\Environment;
use Controller\Exception;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;

class DataSource {

    protected $dataSourceName = 'dataSource';
    /** @var \Service\Manager\Mongo\Manager $organizationManager */
    protected $organizationManager;
    /** @var EntityManager $entityManager */
    protected $entityManager;
    /** @var \Service\SocketClient $socketClient */
    protected $socketClient;
    /** @var ConnectionService $connectionService */
    protected $connectionService;
    /** @var array */
    public $parameterMapping = array(
        'period_start' => 'START_DATE',
        'period_end' => 'END_DATE',
        'filter' => 'STRUCTURE',
        'NRORG_ZMART_PRINCIPAL' => 'NRORG'
    );
    const ARRAY_INCREMENT_DATE_TYPE = array(
        'D' => ' days',
        'M' => ' month',
        'A' => ' year',
    );

    public function __construct(Manager $organizationManager, EntityManager $entityManager, \Service\SocketClient $socketClient, ConnectionService $connectionService) {
        $this->organizationManager = $organizationManager;
        $this->entityManager = $entityManager;
        $this->socketClient = $socketClient;
        $this->connectionService = $connectionService;
    }

    public function findByIndicator($indicator) {
        $filterCriteria = new FilterCriteria('indicator', array(array("columnName"=>"_id", "value" => $indicator, "operator" => "=")));
        $indicator = $this->organizationManager->findBy($filterCriteria)->getRows();
        $dataSource = null;
        if(count($indicator) > 0 && isset($indicator[0]['dataSource'])) {
            $dataSourceRows = $this->findById($indicator[0]['dataSource']);
            if(!empty($dataSourceRows)) {
                $dataSource = $dataSourceRows[0];
            }
        }
        return $dataSource;
    }

    public function checkIndicatorsWithDataSource($dataSources){
        $indicatorService = InstanceManager::getInstance()->getService('\Service\Indicator');
        foreach($dataSources as $dataSource){
            $indicatorWithDataSource = $indicatorService->findByCriteria("dataSource",$dataSource["_id"],"=");
            if(!empty($indicatorWithDataSource)){
                throw new \Exception(Exception::buildErrorJSON(
                    "CANNOT_DELETE_DATASOURCE",
                    array(
                        "indicator" => implode(", ",array_column($indicatorWithDataSource,"name")),
                        "dataSource" => $dataSource["name"]
                    )
                ), 1);
            }
        }
    }

    public function findById($dataSource) {
        $filterCriteria = new FilterCriteria($this->dataSourceName, array(array(
            "columnName"=>"_id",
            "value" => $dataSource,
            "operator" => "="
        )));

        return $this->organizationManager->findBy($filterCriteria)->getRows();
    }

    public function findBy(FilterCriteria $filterCriteria) {
        return $this->organizationManager->findBy($filterCriteria)->getRows();
    }

    protected function getOrderByKeys($orderBy){
        return implode(',', array_keys($orderBy));
    }



    public function setLog($dataBefore,  $id)
    {
        $dataAfter = $this->findById($id);
        Utilities::setLog($dataBefore, $dataAfter, $this->dataSourceName, $id);
    }

    public function setLogIndex($dataBefore, $dataAfter, $id)
    {
        Utilities::setLog($dataBefore, $dataAfter, 'index', $id);

    }

    /**
     * @param $query
     *
     * @return QueryBuilder
     */
    protected function createSelectQuery($query, Connection $customConnection = null, $postConditionsExpression = "",
                                         $orderBy = "", $addAlias = true) {
        if(is_null($customConnection)) {
            $select = $this->entityManager->getConnection()->createQueryBuilder()->select('*');
        } else {
            $select = $customConnection->createQueryBuilder()->select('*');
        }

        if(!empty($postConditionsExpression)){
            $postConditionsExpression = " WHERE {$postConditionsExpression}";
        }

        if(empty($orderBy)){
            $orderByQuery = "";
        }
        else{
            $orderBy = $this->getOrderByKeys($orderBy);
            $orderByQuery = "ORDER BY ".$orderBy;
        }

        $alias = "";
        if($addAlias){
            $alias = "A";
        }

        $query = <<<SQL
          ( $query $alias ) A  $postConditionsExpression $orderByQuery
SQL;
        $select->from($query);
        return $select;
    }


    private function removeIdFromIndexes($indexes){
        foreach($indexes as $i => $index){
            if($index["name"] == "_id_"){
                unset($indexes[$i]);
                return $this->objectToArray(array_values($indexes));
            }
        }
        return $indexes;
    }

    private function objectToArray($array){
        $newArr = [];
        foreach($array as $i => $arr){
            foreach((array)$arr as $a){
                $build = array();
                foreach($a as $key => $value){
                    $build[$key] = $value;
                }
            }
            array_push($newArr,$build);
        }
        return $newArr;
    }

    public function getIndexes($dataSource) {
        $indexes = $this->organizationManager->getIndexInfo($dataSource);
        $indexes = $this->removeIdFromIndexes(iterator_to_array($indexes));
        return $indexes;
    }

    public function getIndexesHistory($dataSource,Manager $manager) {
        $indexes = $manager->getIndexInfo($dataSource);
        $indexes = $this->removeIdFromIndexes(iterator_to_array($indexes));
        return $indexes;
    }

    public function deleteIndexes($rows) {
        foreach($rows as $row){
            $collection = substr($row["ns"], strpos($row["ns"], ".") + 1);
            $this->organizationManager->command(array(
                "deleteIndexes" => $collection,
                "index" => $row["name"]
            ));
        }
    }

    public function saveIndexes($dataSource, $name, $columns, $orders) {
        $keys = [];
        foreach($columns as $i => $column){
            if($column){
                $keys[$column] = (int)$orders[$i];
            }
        }
        $options = array("name" => $name);
        return $this->organizationManager->createIndex($dataSource, $keys, $options);
    }


    protected function parseDefaultQueryParams(FilterCriteria $criteria) {
        $conditions = Utilities::extractConditionInFilter($criteria, array_keys($this->parameterMapping));
        $params = array();
        foreach ($conditions as $condition) {
            if (isset($this->parameterMapping[$condition['columnName']])) {
                $params[$this->parameterMapping[$condition['columnName']]] = $condition['value'];
            }
        }
        return $params;
    }

    /**
     * @param $params
     *
     * @return array
     */
    protected function inferTypes($params) {
        $types = array();
        foreach ($params as $name => $value) {
            $types[$name] = ParameterTypeInferer::inferType($value);
        }
        return $types;
    }

    /**
     * @param FilterCriteria $filterCriteria
     * @param QueryBuilder   $query
     */
    protected function processGroupBy(FilterCriteria $filterCriteria, QueryBuilder $query) {
        $query->addGroupBy($filterCriteria->getGroupBy());
    }

    /**
     * @param FilterCriteria $filterCriteria
     * @param QueryBuilder   $query
     */
    protected function processOrderBy(FilterCriteria $filterCriteria, QueryBuilder $query) {
        foreach ($filterCriteria->getOrderBy() as $columnName => $direction) {
            $query->addOrderBy($columnName, $direction);
        }
    }

    protected function processConditions(FilterCriteria $criteria, QueryBuilder $query) {
        $defaultParams = $this->parseDefaultQueryParams($criteria);
        $this->processGroupBy($criteria, $query);
        $this->processOrderBy($criteria, $query);
        return $defaultParams;
    }

    public function execQuery($query, FilterCriteria $criteria, Connection $customConnection = null) {
        $query = $this->createSelectQuery($query, $customConnection);
        $params = $this->processConditions($criteria, $query);
        $types = $this->inferTypes($params);
        $query->setParameters($params, $types);
        return $query->execute()->fetchAll();
    }

    public function execQueryFromWebSocket($query, FilterCriteria $criteria, $organizationId, $maxWaitTime = null,
                                           $connectionId = null, $postConditionsExpression = "", $orderBy = "",
                                           $addAlias = true,$dataSourceName = 'queryRunny') {
        $query = $this->createSelectQuery($query,null, $postConditionsExpression, $orderBy, $addAlias);
        $params = $this->processConditions($criteria, $query);
        $query = $query->getSql();
        $query = $this->replaceQueryParameters($query, $params);
        $connectionParams = $this->connectionService->getConnectionParams($connectionId);
        $newConnectionParams = array(
            "userName"=> isset($connectionParams['username']) ? $connectionParams['username'] : '',
            "userNameSsl"=> isset($connectionParams['usernameSsl']) ? $connectionParams['usernameSsl'] : '',
            "password" => isset($connectionParams['password']) ? $connectionParams['password'] : '',
            "passwordSsl" => isset($connectionParams['passwordSsl']) ? $connectionParams['passwordSsl'] : '',
            "dbUrlSsl" => isset($connectionParams['dbUrlSsl']) ? $connectionParams['dbUrlSsl'] : $this->buildConnectionStringSsl($connectionParams),
            "dbUrl" => isset($connectionParams['dbUrl']) ? $connectionParams['dbUrl'] : '',
            "driverClass" => $connectionParams['driverClass'],
            "dialect" => $connectionParams['dialect'],
            "oracleWalletUse" => isset($connectionParams['oracleWalletUse']) ? $connectionParams['oracleWalletUse'] : 'N'
        );
        if ($newConnectionParams['oracleWalletUse'] == 'Y') {
            $wallet = $this->getWallet($newConnectionParams);
            if ( Utilities::getOrganizationUseWallet($wallet)) {
                $queryResult = $this->execQueryWallet($query, $dataSourceName, $wallet);
            } else {
                $response = $this->socketClient->execQuery($query, $organizationId, $maxWaitTime ?: 15, $newConnectionParams);
            }
        } else {
            $response = $this->socketClient->execQuery($query, $organizationId, $maxWaitTime ?: 15, $newConnectionParams);
        }

        $response = empty($response)? array() : $response[0];

        if (isset($response['event']) && $response['event'] === SocketClient::NO_CLIENTS_CONNECTED) {
            throw Exception::baseConnectFail();
        } else if (isset($response['hasError']) && $response['hasError']) {
            throw Exception::baseExecuteQueryError($query, new \Exception($response['error']));
        }
        if(empty($queryResult)){
            $queryResult = !empty($response['data']) && !empty($response['data']['response']) ? $response['data']['response'] : array();
        }
        return $queryResult;
    }

    private function buildConnectionStringSsl($connectionParams) {
        $connectionURL = null;
        if(isset($connectionParams['dbname']) && isset($connectionParams['port']) && isset($connectionParams['host']) && isset($connectionParams['dbname'])) {
            $connectionURL = "jdbc:oracle:thin:@".$connectionParams['host'].":".$connectionParams['port']."/".$connectionParams['dbname'];
            $connectionURL = OpensslCrypt::encrypt($connectionURL);
        }
        return $connectionURL;
    }

    public function clearAllRelations($dataSourceId){
        $dataSource = $this->findById($dataSourceId);
        $dataSource[0]["relatedStructuresType"] = [];
        return $dataSource;
    }

    private function buildWhereFromRelations($decodedExternalId, $relations = array()) {
        $conditions = array();
        foreach ($relations as $relation) {
            $key = $relation['dataSourceColumn'];
            $value = $decodedExternalId[$relation['structureColumn']];
            $conditions[] = "{$key}='{$value}'";
        }
        return implode(' AND ', $conditions);
    }

    public function buildWhereClause($structures, $dataSource, $previousConditionExpression = "") {
        $objToReturn = "";
        if(!Utilities::isNumericArray($structures)) {
            $structures =  array($structures);
        }
        if (isset($dataSource['relatedStructuresType']) && !empty($structures)) {
            $whereStructure = array();
            foreach ($structures as $structure) {
                $decodedExternalId = Utilities::decodeStructureExternalId($structure['externalId']);
                //$decodedExternalId = $structure['externalId']; //@TODO When all DataBases start use object externalId, uncomment this line
                foreach ($dataSource['relatedStructuresType'] as $relatedStructureType) {
                    if ($relatedStructureType['structureType'] === $structure['type']) {
                        $objToReturn = $this->buildWhereFromRelations($decodedExternalId, $relatedStructureType['relations']);
                    }
                }
                $whereStructure[] = $objToReturn;
            }
            $objToReturn = '(('.implode(' ) OR ( ',$whereStructure).'))';
        }

        $objToReturn = !empty($previousConditionExpression) ? $objToReturn.' AND '.$previousConditionExpression : $objToReturn;

        return $objToReturn;
    }

    public function replaceQueryParameters($query, $replaceMapping = array()) {
        foreach ($this->parameterMapping as $key => $value) {
            $param = ':' . $value;
            if (strpos($query, $param) !== false) {
                $query = str_replace($param, $this->getQueryValue($replaceMapping, $value), $query);
            }
        }
        return $query;
    }

    protected function getQueryValue($replaceMapping, $key) {
        $value = "''";
        if (isset($replaceMapping[$key])) {
            $value = is_string($replaceMapping[$key])? "'{$replaceMapping[$key]}'" : $replaceMapping[$key];
        }

        return $value;
    }

    public function getColumnsFromOriginDataSources($originDataSource, &$columns){
        if(!$originDataSource){
            throw Exception::syncJoinDataSourceColumnsError();
        }

        $originDataSourceName = $originDataSource['name'];
        foreach ($originDataSource['columns'] as $columnToImport) {
            if(isset($columnToImport['isGenerated']) && $columnToImport['isGenerated']){
                $duplicatedColumn = false;
                foreach ($columns as &$column) {
                    if($column['name'] === $columnToImport['name']){
                        $column['columnOriginDataSources'] = $column['columnOriginDataSources'] . ", " . $originDataSourceName;
                        $duplicatedColumn = true;
                    }
                }

                if(!$duplicatedColumn){
                    if (isset($columnToImport['isFact']) && $columnToImport['isFact']){
                        $columnToImport['joinFunction'] = isset($columnToImport['function']) ? $columnToImport['function'] : "SUM";
                    }
                    $columnToImport['columnOriginDataSources'] = $originDataSourceName;
                    array_push($columns, $columnToImport);
                }
            }
        }
    }

    public function dataSourceGeneratedData($values, $dataSourceId){
        foreach ($values as $key => $value) {
            $baseIncrement = '';
            $valuebase = $value[0];
            $typeIncrement = $value[1];
            $zmartFilterCriteria = new ZmartFilterCriteria($dataSourceId);
            $zmartFilterCriteria->addOrderBy('_id');
            $zmartFilterCriteria->setGroupBy(array($key));
            $zmartFilterCriteria->setPageSize(999999999);
            $rows = $this->organizationManager->aggregate($zmartFilterCriteria)->getRows();
            foreach ($rows as $row) {
                $update = $this->getValueUpdate($valuebase, $row['_id'], $typeIncrement, $baseIncrement);
                $filter = array($key => $row['_id']);
                $updateMongo = array('$set' => array($key => $update));
                $this->organizationManager->mongo->update($dataSourceId, $filter, $updateMongo, false, true);
            }
        }
    }

    protected function getWallet(array $newConnectionParams)
    {
        return Utilities::getWallet($newConnectionParams);
    }

    /**
     * @param $query
     * @param $dataSourceName
     * @param $wallet
     * @return array
     */
    protected static function execQueryWallet($query, $dataSourceName, $wallet): array
    {
        return  Utilities::execQueryWallet($query, $dataSourceName, $wallet);
    }

    /**
     * @param $valuebase
     * @param $row
     * @return mixed
     */
    public function getValueUpdate($valuebase, &$valueDB, $typeIncrement, &$baseIncrement)
    {
        $update = $valueDB;
        if ($typeIncrement[0] == '-' || $typeIncrement[0] == '+') {
            if(is_object($valueDB) && get_class($valueDB) == 'DateTime') {
                $type = self::ARRAY_INCREMENT_DATE_TYPE[$typeIncrement[1]];
                $operator = $typeIncrement[0];
                /** @var $valueDB \DateTime */
                $update = clone $valueDB;
                $update->modify($operator . ' ' . $valuebase . $type);
                $update = Type::getType('date')->convertToDatabaseValue($update);
                $valueDB = Type::getType('date')->convertToDatabaseValue($valueDB);
            }
        } else {
            $baseIncrement = $this->setInitBaseIncrement($typeIncrement, $baseIncrement);
            $update = $valuebase . $baseIncrement;
            $update = $this->setTypeEqualBase($valueDB, $update);
        }
        return $update;
    }

    /**
     * @param $typeIncrement
     * @param $baseIncrement
     * @return int|string
     */
    public function setInitBaseIncrement($typeIncrement, &$baseIncrement)
    {
        if (!$baseIncrement) {
            if ($typeIncrement == 'NUMERIC') {
                $baseIncrement = 1;
            } else if ($typeIncrement == 'ALPHABETIC') {
                $baseIncrement = 'A';
            }
        } else {
            $baseIncrement++;
        }
        return $baseIncrement;
    }

    /**
     * @param $valueDB
     * @param string $update
     * @return float|int|string
     */
    public function setTypeEqualBase($valueDB, string $update)
    {
        if (is_int($valueDB)) {
            $update = intval($update);
        } else if (is_float($valueDB)) {
            $update = floatval($update);
        }
        return $update;
    }
}
