<?php


namespace Service;


use Controller\Exception;
use MongoDB\BSON\UTCDateTime;
use Service\Manager\Mongo\Manager;
use Service\PeriodMathExpression\Period;
use Util\InstanceProvider;
use Util\Utilities;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;

class QueryRunnerMining
{

    const NSP        = 'ZmartDataCollector';
    const SERVER_KEY = 'you can create!';
    const PING_ROOM  = 'PING_ROOM';

    const COLLECTION_NAME = "importHistory";
    const INDEX_NAME = "indexCreateDate";

    // Events
    const CREATE_NSP         = 'createNamespace';
    const CREATE_ROOM        = 'createRoom';
    const JOIN_ROOM          = 'joinRoom';
    const EMIT               = 'emit';
    const PING               = 'pingConnection';
    const START_IMPORTATION  = 'startImportation';
    const FINISH_IMPORTATION = 'finishImportation';
    const NOTIFY_ERROR       = 'onError';
    const UPDATE_IMPORT_PROCESS = 'updateImportProcess';
    const GET_CREATOR_DATASOURCE = 'getCreatorDatasource';

    /** @var \Service\Manager\Mongo\Manager $organizationManager */
    protected $organizationManager;
    /** @var \Service\SocketClient $socketClient */
    protected $socketClient;
    /** @var \Service\Manager\Mongo\Manager $defaultOrgManager */
    protected $defaultOrgManager;

    public function __construct (Manager $organizationManager, \Service\SocketClient $socketClient, Manager $defaultOrgManager) {
        $this->organizationManager = $organizationManager;
        $this->socketClient = $socketClient;
        $this->defaultOrgManager = $defaultOrgManager;
    }

    public function executeQueryMining ($connectionParams, $maxWaitTime = null){

        $environment = Utilities::getEnvironment();
        $dataTimeString = Utilities::getTimezonedNow()->format('YmdHis');
        $baseProcess = $environment->getCurrentOrganizationId().'_'.$environment->getCurrentUserId();

        $organization = Utilities::getOrganizationIdForMongo();
        $process = $baseProcess.'_'.$dataTimeString;
        $process = str_pad($process, 47, "0", STR_PAD_LEFT);
        $query = $connectionParams['query'];
        $queryKeys = $connectionParams['keys'];
        $pageSize = 5000;
        $dbUrl = $connectionParams['dbUrlSsl'];
        $userName = null;
        $password = null;
        $scriptBefore = "";
        $scriptAfter = "";
        $execAfterIfImportError = false;
        $execAfterIfBeforeScriptError = false;
        $dbUrlSsl = $connectionParams['dbUrlSsl'];
        $usernameSsl = $connectionParams['usernameSsl'];
        $passwordSsl = $connectionParams['passwordSsl'];
        $clientRoom = $connectionParams['room'];
        $driverClass = $connectionParams['driverClass'];
        $dialect = $connectionParams['dialect'];
        $connectionId = $connectionParams["connectionId"];

        $historyItem = $this->factoryHistoryItem($process, $organization, $baseProcess, 0, $query, $queryKeys,$connectionId);
        $dataSet =  new DataSet(self::COLLECTION_NAME, array($historyItem));
        $this->organizationManager->persist($dataSet);

        $this->socketClient->startImportation($organization, $process, $query, $queryKeys, $pageSize, $dbUrl, $userName, $password,
            $scriptBefore, $scriptAfter, $execAfterIfImportError, $execAfterIfBeforeScriptError,
            $clientRoom, $dialect, $driverClass, $usernameSsl, $passwordSsl, $dbUrlSsl);
        return array(array('PROCESS' => $process));
    }

    protected function factoryHistoryItem($process, $organizationId, $baseProcess, $subprocesses, $query, $queryKeys,$connectionId,$logs = array(), $importStatus = "start") {
        return array(
            'process'        => $process,
            'organizationId' => $organizationId,
            'baseProcess'    => $baseProcess,
            'pageSize'       => null,
            'subprocesses'   => $subprocesses,
            'imported'       => 0,
            'logs'           =>  $logs,
            'hasError'       => false,
            'createDate'     => new UTCDateTime(),
            'endDate'        => null,
            'elapsedTime'    => null,
            'locked'         => false,
            'query'          => $query,
            'queryKeys'      => $queryKeys,
            'connectionId'   => $connectionId,
            'importStatus'   => $importStatus
        );
    }

    public function getResult($condition,$processQueryRunner){

        $filterCriteria = new ZmartFilterCriteria($processQueryRunner, $condition);

        return $this->organizationManager->findBy($filterCriteria)->getRows();
    }

    public function clearImportHistory(){
        $this->dropAllHistoryOdd();
    }

    public function dropAllHistoryOdd(){
        $allOrganizationZmart = InstanceProvider::getLogImportation()->getAllOrganizationZmart();
        $total = count($allOrganizationZmart);
        echo("Total de $total clientes que tem o zmart \n");
        $start= microtime(true);
        foreach ($allOrganizationZmart as $organization) {
            try {
                $nrorg = $organization['NRORG'];
                $manager = LogImportation::buildManager($nrorg);
                $this->dropHistoryOdd($nrorg,$manager);
                $this->createdIndexHistory($nrorg,$manager);
                $this->dropCollectionQueryRunnerFromScipt($nrorg,$manager);
            }catch (\Exception $e){
                var_dump($organization);
                $message = $e->getMessage();
                $nrorg = $organization['NRORG'];
                echo("Erro no cliente $nrorg : $message \n");
            }
        }
        $end = microtime(true);
        $totalTime = $end - $start;

        echo("Tempo gasto em toda oa operação foi de $totalTime \n");
    }

    public function dropCollectionQueryRunnerFromScipt($nrorg,$manager){
        $collections = $this->dropCollectionQueryRunner($nrorg,$manager);
        $count = count($collections['collectionDrop']);
        $date = $collections['date'];
        echo("Foi removido $count collections de query runner com data anterior a $date");
    }

    public function createdIndexHistory($nrorg, Manager $manager){
        $indexes = InstanceProvider::getDataSource()->getIndexesHistory(self::COLLECTION_NAME,$manager);
        $indexes = Utilities::getArrayFromAssociativeArray($indexes,'name');
        if(!in_array(self::INDEX_NAME,$indexes)){
            echo("Iniciando criação de index  do cliente $nrorg \n");
            $keys = array(
                "createDate" => -1
            );
            $options = array("name" => self::INDEX_NAME);
            $manager->createIndex(self::COLLECTION_NAME, $keys, $options);
            echo("Index criado  do cliente $nrorg \n");
        }
    }

    public function dropHistoryOdd($nrorg, Manager $manager){

        echo("Iniciando delete de logs de importaçẽos antigos do cliente $nrorg \n");
        $start= microtime(true);
        list($dataTimeString, $condition) = $this->createConditionDeleteLog();
        $filterCriteria = new ZmartFilterCriteria(self::COLLECTION_NAME, $condition);
        $deletedCount = $manager->remove(self::COLLECTION_NAME,$filterCriteria)->getDeletedCount();
        $end = microtime(true);
        $totalTime = $end - $start ;
        echo("Foram deletadas $deletedCount logs de importações do cliente $nrorg feitas anterior a $dataTimeString, tempo gasto $totalTime \n");

    }

    public function dropCollectionQueryRunner($nrorg = null ,Manager $manager = null){
        $collectionDrop = array();
        $environment = Utilities::getEnvironment();
        if(!$nrorg){
            $cdoperador = $environment->getCurrentUserId();
            $nrorg = $environment->getCurrentOrganizationId();
            $notValidOperador = false;
        }else{
            $cdoperador = null;
            $notValidOperador = true;

        }
        if(!$manager){
            $manager = $this->organizationManager;
        }

        $collectionNames = $manager->getCollectionNames();
        $organizationId = str_pad($nrorg,19, '0',STR_PAD_LEFT);
        $timeLife = Utilities::getTimeLifeCollectionQueryRunner();
        $date = Period::getDate("-$timeLife");
        $dataTimeString = Utilities::staticDateInverse($date);
        $dataTimeString .='235959';

        foreach ($collectionNames as $item) {
            if(strlen($item) == 47){
                $parts = explode('_',$item);
                if(!empty($parts[2]) &&
                    !empty($parts[1]) &&
                    !empty($parts[0]) &&
                    ($parts[1] === $cdoperador || $notValidOperador) &&
                    strlen($parts[2]) === 14 &&
                    $parts[0] === $organizationId &&
                    $dataTimeString > $parts[2]){
                    $manager->dropCollection($item);
                    $collectionDrop[] = $item;
                }
            }
        }
        return array("collectionDrop" => $collectionDrop,'date' => $date);
    }

    public function getColumnsMetaData($connectionParams){
        $query = $connectionParams['query'];
        $clientRoom = $connectionParams['room'];
        $connectionParams['dbUrl'] = $connectionParams['dbUrlSsl'];
        $newConnectionParams = $this->buildConnectionParams($connectionParams);
        $response = $this->socketClient->getColumnsMetaData($query,$clientRoom,120,$newConnectionParams);
        $response = empty($response)? array() : $response[0];

        if (isset($response['event']) && $response['event'] === SocketClient::NO_CLIENTS_CONNECTED) {
            throw Exception::connectFail();
        } else if (isset($response['hasError']) && $response['hasError'] === true) {
            throw Exception::executeQueryError($query, new \Exception($response['error']));
        }
        return !empty($response['data']) && !empty($response['data']['response'])? $response['data']['response'] : array();
    }

    public function buildConnectionParams($connectionParams){
        return array(
            "userNameSsl"	  => isset($connectionParams["usernameSsl"])    ? $connectionParams["usernameSsl"]    : null,
            "dialect"	  => isset($connectionParams["dialect"])     ? $connectionParams["dialect"]     : null,
            "dbUrlSsl"    => isset($connectionParams['dbUrlSsl'])    ? $connectionParams['dbUrlSsl']    :null,
            "passwordSsl" => isset($connectionParams['passwordSsl']) ? $connectionParams['passwordSsl'] :null,
            "usernameSsl" => isset($connectionParams['usernameSsl']) ? $connectionParams['usernameSsl'] :null,
            "driverClass" => isset($connectionParams["driverClass"]) ? $connectionParams["driverClass"] : null
        );
    }

    /**
     * @return array
     * @throws \Exception
     */
    private function createConditionDeleteLog(): array
    {
        $timeLifeLogImport = Utilities::getTimeLifeLogImport();
        $dataTimeString = Period::getDate("-$timeLifeLogImport");
        $condition = array(
            array(
                "columnName" => "createDate",
                "value" => Utilities::convertStringDataBrToMongoDate($dataTimeString, ' 23:59:59'),
                "operator" => ZmartFilterCriteria::LTE
            )
        );
        return array($dataTimeString, $condition);
    }
}
