/**
 *
 * EventEngine is a service that provides events handlers, builder and management of business and DOM events.
 */
function EventEngine(eventFactory, eventHandler, templateManager, ScreenService, NavigationLog) {
	var self = this;
	/**
	 * @constant key-value constant that binds a Zeedhi event name to its
	 *           equivalent on its origin or its name on the object.
	 */
	var eventNameToEventFunction = {
		'WidgetBeforeInlineEditMoveRow': 'beforeInlineEditMoveRow',
		'WidgetBeforeMoveRow': 'beforeMoveRow',
		'WidgetAfterMoveRow': 'afterMoveRow',
		'WidgetBeforeNewRow': 'beforeNewRow',
		'WidgetNewRow': 'onNewRow',
		'WidgetOnEnter': 'onEnter',
		'WidgetOnEnterOnce': 'onEnterOnce',
		'WidgetOnEnterTab': 'onEnterTab',
		'WidgetOnSync': 'onSync',
		'WidgetBeforeSelectRow': 'beforeSelectRow',
		'WidgetAfterSelectRow': 'afterSelectRow',
		'WidgetBeforeUnselectRow': 'beforeUnselectRow',
		'WidgetAfterUnselectRow': 'afterUnselectRow',
		'WidgetOnCancel': 'onCancel',
		'WidgetOnSave': 'onSave',
		'WidgetBeforeSelectAll': 'beforeSelectAll',
		'WidgetAfterSelectAll': 'afterSelectAll',
		'WidgetBeforeUnselectAll': 'beforeUnselectAll',
		'WidgetAfterUnselectAll': 'afterUnselectAll',
		'WidgetOnInitDataSource': 'initDataSource',
		'WidgetAfterLoadDataSource': 'afterLoadDataSource',
		'WidgetAfterLoadDataSourceError': 'afterLoadDataSourceError',
		'WidgetPopUpOnCancel': 'popupOnCancel',
		'WidgetOnActivate': 'onActivate',
		'WidgetOnDeactivate': 'onDeactivate',
		'WidgetOnDateClick': 'dateClick',
		'WidgetOnMonthClick': 'monthClick',
		'WidgetOnYearClick': 'yearClick',
		'WidgetOnDashboardClick': 'onDashboardClick',
		'WidgetOnDashboardConfigClick': 'onDashboardConfigClick',
		'WidgetOnSaveInlineEdit': 'onSaveInlineEdit',
		'WidgetOnCancelInlineEdit': 'onCancelInlineEdit',
		'WidgetOnSaveRelations': 'onSaveRelations',
		'WidgetOnOpenDetail': 'onOpenDetail',
		'WidgetOnCloseSwipe': 'onCloseSwipe',
		'WidgetOnNewColor': 'onNewColor',
		'WidgetOnChangeMonth': 'onChangeMonth',
		'WidgetOnChangeYear': 'onChangeYear',
		'WidgetOnChangeDecade': 'onChangeDecade',

		'ContainerOnBack': 'onBack',
		'ContainerBeforeClose': 'beforeClose',
		'ContainerAfterClose': 'afterClose',
		'ContainerBeforeinit': 'beforeinit',
		'ContainerAfterinit': 'afterinit',
		'ContainerAfterOpenreport': 'afterOpenreport',

		'WizardOnDecision': 'onDecision',
		'WizardAfterOpenStage': 'afterOpenStage',
		'WizardBeforePreviousStage': 'beforePreviousStage',
		'WizardAfterPreviousStage': 'afterPreviousStage',
		'WizardBeforeNextStage': 'beforeNextStage',
		'WizardAfterNextStage': 'afterNextStage',
		'WizardOnFinish': 'onFinish',
		'WizardOnCancel': 'onCancel',

		'FieldOnChange': 'change',
		'FieldOnInput': 'input',
		'FieldOnCancel': 'onCancel',
		'FieldOnAdd': 'onAdd',
		'FieldOnConfirm': 'onConfirm',
		'FieldOnClick': 'touchstart',
		'FieldOnBlur': 'blur',
		'FieldOnKeypress': 'keypress',
		'FieldOnKeyup': 'keyup',
		'FieldOnKeydown': 'keydown',
		'FieldAfterMoveRow': 'afterMoveRow',
		'FieldBeforeMoveRow': 'beforeMoveRow',
		'FieldOnFocus': 'focus',
		'FieldBeforeSelectOpen': 'beforeSelectOpen',
		'FieldAfterSelectOpen': 'afterSelectOpen',
		'FieldAfterSelectFile': 'afterSelectFile',
		'FieldAfterDeselectFile': 'afterDeselectFile',
		'FieldInlineEdit': 'onInlineEdit',
		'FieldOnNewColor': 'onNewColor',

		'ActionEvent': 'click',

		'ElementOnClick': 'click',
		'ElementOnTouchstart': 'touchstart',

		'NotificationEvent': 'click',

		'onUpdate': 'onUpdate',

		'RepositoryOnRetry': 'onRetry',
		'EventOnContainerOpen': 'onContainerOpen',
		'EventOnRequestMetaDataStart': 'onRequestMetaDataStart',
		'EventOnRequestStart': 'onRequestStart',
		'EventOnRequestEnd': 'onRequestEnd',
		'EventOnRequestRetry': 'onRequestRetry',
		'EventOnRequestSuccess': 'onRequestSuccess',
		'EventOnRequestError': 'onRequestError',
		'EventOnRequestIdle': 'onRequestIdle',

		'KanbanBeforeMoveItem': 'beforeMoveItem',
		'KanbanAfterMoveItem': 'afterMoveItem',
		'KanbanOnTaskClick': 'onTaskClick',

		'MenuOnClick': 'click',

		'ProductOnClick': 'onProductClick'
	};

	var SUFFIX_LISTENER = 'listener';

	var stopDispatchEvent = {
		stopped: false,
		reason: undefined
	};
	//fallback to browsers in pc
	if (!Zepto.os.phone && !Zepto.os.tablet) {
		eventNameToEventFunction.FieldOnClick = "click";
		eventNameToEventFunction.ElementOnTouchstart = "click";
	}
	var eventsWhenNotUpdateTemplate = ['onUpdate'];

	//These auxiliary functions are used within the for cycle.
	//This function is called in the case 'FieldOnTap'
	function eventResolverFactory(eventWidget) {
		return function eventResolver(e) {
			eventWidget.notify(e);
			templateManager.updateTemplate();
		};
	}

	/**
	 * A Sugar method to bind a single event in owner
	 * @returns Number - the index of the created listener
	 */
	this.bindEvent = function(owner, eventName, code, dataSourceCallback, rowCallback, selectedRowCallback, rowGroupCallback) {
		this.bindEvents(owner, [{
			name: eventName,
			code: code
		}], dataSourceCallback, rowCallback, selectedRowCallback, rowGroupCallback, true);
		var ownerListeners = owner[eventNameToEventFunction[eventName] + SUFFIX_LISTENER];
		return ownerListeners ? ownerListeners.length - 1 : -1;
	};
	this.bindEvents = function(owner, events, dataSourceCallback, rowCallback, selectedRowCallback, rowGroupCallback, forceBind) {
		if (!owner.isBindedEvent || forceBind) {
			var eventCallback = function(e, eventData) {
				if (!stopDispatchEvent.stopped && !eventData.isPropagationStopped()) {
					return eventHandler.notify(e, {
						"e": eventData,
						"owner": owner,
						"dataSource": dataSourceCallback ? dataSourceCallback() : [],
						"row": rowCallback ? rowCallback() : {},
						"selectedRow": selectedRowCallback ? selectedRowCallback() : {},
						"rowGroup": rowGroupCallback ? rowGroupCallback() : {}
					});
				}
			};

			for (var i = 0; i < events.length; i++) {
				var currentEvent = events[i];
				var eventObj = eventFactory.factory(currentEvent.name, currentEvent.id, {
					"owner": owner,
					event: currentEvent
				}, currentEvent.code, currentEvent.serviceName, currentEvent.requestType);
				var eventWidget = new EventWidget(eventObj, eventCallback);
				switch (currentEvent.name) {
					case 'FieldOnTap':
						owner.on('tap', eventResolverFactory(eventWidget));
						break;
					case 'FieldOnLongTap':
						owner.on('longtap', eventResolverFactory(eventWidget));
						break;
					default:
						treatDefaultEvent(currentEvent, owner, eventWidget);
						break;
				}
			}
			owner.isBindedEvent = true;
		}
	};

	function bindBusinessEvent(owner, eventName, fn) {
		if (owner.field && !owner.field.bindedEvents) {
			owner.field.bindedEvents = [];
		}
		if (owner.field && owner.field.bindedEvents.indexOf(eventName) === -1) {
			if (!owner.field[eventName]) {
				self.createEvent(owner.field, eventName);
			}
			owner.field[eventName](fn);
			owner.field.bindedEvents.push(eventName);
		}
	}
	var updateTemplate = function(methodName) {
		if (eventsWhenNotUpdateTemplate.indexOf(methodName) === -1) {
			templateManager.updateTemplate();
		}
	};
	var addListener = function(owner, methodName, eventWidget) {
		owner[methodName](function(e) {
			var result = eventWidget.notify(e);
			updateTemplate(methodName);
			return result;
		});
	};
	var hasAnActiveField = function(owner) {
		return owner.field && !owner.field.aclReadOnly && !owner.field.readOnly;
	};
	var thereIsNoField = function(owner) {
		return !owner.field;
	};
	var shouldIgnoreEvent = function(e) {
		return typeof e.__ignore__ !== "undefined";
	};
	var buildJQueryEvent = function(owner, methodName, currentEvent, eventWidget) {
		return function(evt) {
			if ((!evt || !shouldIgnoreEvent(evt)) && (thereIsNoField(owner) || hasAnActiveField(owner))) {
				if (currentEvent.name === 'FieldOnClick' && !(owner.shouldNotPreventDefaultOnClick || (owner.field && owner.field.shouldNotPreventDefaultOnClick))) {
					evt.preventDefault();
				}
				eventWidget.notify(evt);
				updateTemplate(methodName);
			}
		};
	};

	function treatDefaultEvent(currentEvent, owner, eventWidget) {
		if (self.hasEvent(currentEvent.name)) {
			var methodName = eventNameToEventFunction[currentEvent.name];
			if (owner.off) {
				var fn = buildJQueryEvent(owner, methodName, currentEvent, eventWidget);
				bindBusinessEvent(owner, methodName, fn);
				owner.on(methodName, fn);
			} else if (owner[methodName]) {
				addListener(owner, methodName, eventWidget);
			} else {
				self.createEvent(owner, methodName);
				addListener(owner, methodName, eventWidget);
			}
		} else {
			throw new Error(ScreenService.i18n("Event __eventName__ not mapped!", {
				eventName: currentEvent.name
			}));
		}
	}

	this.createEvent = function createEvent(owner, eventName, bindFirst) {
		var eventObj = new Eventable();
		owner[eventName] = function(eventCode, params) {
			eventObj.resetAllProperties();

			if (!owner[eventName + SUFFIX_LISTENER]) {
				owner[eventName + SUFFIX_LISTENER] = [];
			}
			if (eventCode && Util.isFunction(eventCode)) {
				if (bindFirst) {
					owner[eventName + SUFFIX_LISTENER].unshift(eventCode);
				} else {
					owner[eventName + SUFFIX_LISTENER].push(eventCode);
				}
			} else {
				var listeners = owner[eventName + SUFFIX_LISTENER];
				if (eventName != "onUpdate") {
					NavigationLog.logEvent(eventName, owner);
				}
				for (var i = 0; i < listeners.length; i++) {
					var eventData = eventCode ? eventCode : params;
					if (!eventData) {
						eventData = {};
					}
					eventData = Util.extend(eventObj, eventData, true);
					eventObj.data = listeners[i](eventData);
				}
			}
			return eventObj;
		};
	};

	this.registerEvent = function(eventName, eventFunctionName) {
		if (eventName && eventFunctionName) {
			if (!this.hasEvent(eventName)) {
				eventNameToEventFunction[eventName] = eventFunctionName;
			} else {
				throw new Error(ScreenService.i18n("Event __eventName__ is already mapped!", {
					eventName: eventName
				}));
			}
		} else {
			throw new Error(ScreenService.i18n("Invalid parameters"));
		}
	};

	this.unregisterEvent = function(eventName) {
		if (eventName) {
			delete eventNameToEventFunction[eventName];
		}
	};

	/**
	 * Unbinds an event from the owner listeners
	 *
	 * @param owner The owner of the event
	 * @param eventName The name of the event
	 * @param eventIndex The index of the listener to be removed
	 * @returns Boolean true if the eventIndex was found and removed, otherwise false
	 * @example eventEngine.unbindEvent(widget, WidgetOnEnter, 5)
	 */
	this.unbindEvent = function unbindEvent(owner, eventName, eventIndex) {
		var successfullyRemoved = false;
		var methodName = eventNameToEventFunction[eventName];
		var listeners = owner[methodName + SUFFIX_LISTENER];

		if (listeners && listeners[eventIndex]) {
			listeners.splice(eventIndex, 1);
			successfullyRemoved = true;
		}

		return successfullyRemoved;
	};

	this.stopDispatchEvent = function(reason) {
		if (!reason) {
			throw new Error(ScreenService.i18n("Reason is mandatory to stop dispatch event"));
		}
		stopDispatchEvent = {
			stopped: true,
			reason: reason
		};
	};

	this.restoreDispatchEvent = function() {
		stopDispatchEvent = {
			stopped: false,
			reason: undefined
		};
	};

	this.isDispatchEventStopped = function() {
		return stopDispatchEvent.stopped;
	};

	this.hasEvent = function(name) {
		return eventNameToEventFunction.hasOwnProperty(name);
	};
}