"use strict";

module.exports = function (grunt) {

    var CONFIG_FILE_PARAM = 'configFile';
    var SRC_FOLDERS_PARAM = 'srcFolders';
    var OPTIONS_PARAM     = 'options';
    var DELIMITER         = '<>';
    var TEMPLATE_SUFFIX   = '.template';
    var DEFAULT_RUN_MODE  = 'development';

    var requiredParameters = [CONFIG_FILE_PARAM, SRC_FOLDERS_PARAM];

    var globalConfig;
    var runModeConfig;

    function getProjectProperties(configFile) {
        return grunt.file.readJSON(configFile);
    }

    function getTemplates(srcFolders) {
        var options = {
            dot: true
        };

        return grunt.file.expand(options, srcFolders.map(function(folder) {
            if (folder.charAt(folder.length-1) !== '/') {
                folder = folder+'/';
            }

            return folder + '*' + TEMPLATE_SUFFIX;
        }));
    }

    function generateFiles(properties, templates, options) {
        var missingProperties = [];
        templates.forEach(function(template) {
            var newFile = generateFile(properties, template);
            grunt.file.write(removeTemplateSuffix(template), newFile);
            missingProperties = missingProperties.concat(getMissingProperties(newFile, template));
        });

        var missingPropertiesCount = arrayCountUniqueValues(missingProperties);
        if (missingPropertiesCount > 0) {
            var errorMsg = 'Missing ' + missingPropertiesCount + ' ' + grunt.util.pluralize(missingPropertiesCount, 'property/properties') + ' on ' + CONFIG_FILE_PARAM + '.';

            if (options && options.force) {
                grunt.log.error(errorMsg);
            } else {
                grunt.fail.warn(errorMsg);
            }
        }
    }

    function generateFile(properties, template) {
        var templateContent = grunt.file.read(template);
        for (var key in properties) {
            var regex = new RegExp(DELIMITER + key + DELIMITER, 'g');
            templateContent = templateContent.replace(regex, properties[key]);
        }
        return templateContent;
    }

    function getMissingProperties(newFile, templateName) {
        var regex = new RegExp(DELIMITER + '(.+?)' + DELIMITER, 'g');
        var missingProperties = newFile.match(regex);

        if (!Array.isArray(missingProperties)) {
            missingProperties = [];
        }

        missingProperties.forEach(function(missingProperty) {
            grunt.log.error('Undefined property "' + removeDelimiters(missingProperty) + '" used on file "' + templateName + '".');
        });

        return missingProperties;
    }

    function removeDelimiters(missingProperty) {
        return missingProperty.substring(DELIMITER.length, missingProperty.length - DELIMITER.length);
    }

    function removeTemplateSuffix(templateName) {
        return templateName.substring(0, templateName.length - TEMPLATE_SUFFIX.length);
    }

    function arrayCountUniqueValues(arr) {
        var uniques = [];
        arr.forEach(function(value) {
            if (uniques.indexOf(value) === -1) {
                uniques.push(value);
            }
        });

        return uniques.length;
    }

    function initConfigValues(runMode, taskName) {
        globalConfig = grunt.config(taskName);
        runModeConfig = getRunModeConfiguration(runMode);
    }

    function getParameter(parameterName, validate) {
        var parameterValue = runModeConfig[parameterName] || globalConfig[parameterName];

        if (!parameterValue && requiredParameters.indexOf(parameterName) !== -1) {
            grunt.fail.warn('Missing parameter "' + parameterName + '".');
        }

        if (!validate(parameterValue)) {
            grunt.fail.warn('Invalid "' + parameterName + '" parameter.');
        }

        return parameterValue;
    }

    function getRunModeConfiguration(runMode) {
        runMode = runMode || DEFAULT_RUN_MODE;
        var runModeConfig = globalConfig[runMode];

        if (runModeConfig === undefined && runMode !== DEFAULT_RUN_MODE) {
            grunt.fail.warn('Configuration for "' + runMode + '" not found.');
        }

        return runModeConfig || {};
    }

    function isStringNotEmpty(str) {
        return typeof str === 'string' && str.length !== 0;
    }

    function isObjectOrUndefined(value) {
        return value === undefined || typeof value === 'object';
    }

    function getCommandLineProperties(args, properties) {
        args.forEach(function(arg) {
            var argParts = arg.split('=');
            var name = argParts[0].substr(2, argParts[0].length - 1);

            if (argParts.length !== 2) {
                throw 'Missing value for command line property "'+name+'"';
            }

            var value = argParts[1];

            properties[name] = value;
        });

        return properties;
    }

    grunt.registerTask('buildEnvironment', function(runMode) {
        try {
            initConfigValues(runMode, this.name);
            var configFile = getParameter(CONFIG_FILE_PARAM, isStringNotEmpty);
            var srcFolders = getParameter(SRC_FOLDERS_PARAM, Array.isArray);
            var options = getParameter(OPTIONS_PARAM, isObjectOrUndefined);

            var parameters = grunt.option.flags();
            var projectProperties = getCommandLineProperties(parameters, getProjectProperties(configFile));
            var templates = getTemplates(srcFolders);

            grunt.log.ok(templates.length + ' template' +
                grunt.util.pluralize(templates.length, '/s') + ' found.');

            generateFiles(projectProperties, templates, options);
        } catch (e) {
            grunt.fail.warn(e);
        }
    });

};