<?php

namespace Services;

use Util\Utilities;
use Util\InstanceProvider;
use Util\RepositoryInstance;
use Exception\Login as Exception;
use Teknisa\Libs\Service\Menu;
use Teknisa\Libs\Service\Api as LibApiService;
use Teknisa\Libs\Service\Login as LibLoginService;
use Teknisa\Libs\Util\Utilities as LibUtilities;
use Teknisa\Libs\Util\ConnectionCustom as LibConnectionCustom;
use Teknisa\Libs\Util\Repositories\UserData as LibUserDataRepository;

class Login
{
    const IS_ACTIVE_POSITIVE = 'Y';
    const DATE_TYPE_DAY = 'D';
    const ID_RESET_PASSWORD = 'R';
    const USES_ACTIVE_DIRECTORY_POSITIVE = 'Y';
    const USES_PASSWORD_CONTROL_POSITIVE = 'Y';
    const USES_IP_RESTRICTION_POSITIVE = 'Y';
    const RECURRENCE_DAYS = 7;

    const BLOCKED_PASSWORD_STATUS = 'B';
    const INVALID_PASSWORD_STATUS = 'I';

    const RESET_PASSWORD_IDENTIFICATION   = 'RESET_PASSWORD';
    const EXPIRED_PASSWORD_IDENTIFICATION = 'EXPIRED_PASSWORD';
    const PASSWORD_WILL_EXPIRE_IDENTIFICATION = 'PASSWORD_WILL_EXPIRE';
    const PASSWORD_EXPIRE_TODAY_IDENTIFICATION = 'PASSWORD_EXPIRE_TODAY';

    protected $loginUrl;
    protected $loginClientSecret;

    protected $libClientId;
    protected $libClientSecret;

    protected $licenseExceptions;

    public function __construct() {
        $config = LibUtilities::getLoginConfig();
        $this->loginUrl          = $config['loginUrl'];
        $this->loginClientSecret = $config['loginClientSecret'];

        $integration = LibUtilities::getLoginIntegration();
        $this->libClientId     = $integration['clientId'];
        $this->libClientSecret = $integration['clientSecret'];

        $this->licenseExceptions = array(
            22, /** Product is not available */
            23, /** Menu items not found */
            24  /** Product date expired */
        );
    }

    /**
     * @param $user
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @return array
     * @throws \Exception
     */
    public function login($user, $password, $productId, $requesterUrl, $attempts) {
        $userData       = InstanceProvider::getLibServiceApi()->getUserData($user, $password);
        $organizationId = $this->getOrganizationId($userData);
        $teknisaOrgId   = $this->getTeknisaOrganizationId($userData);
        $userId         = $userData['USER']['ID'];
        $userName       = $userData['USER']['NAME'];
        $nrParcNegocio  = $userData['USER']['NRPARCNEGOCIO'];

        $userData['ACCESS_PROFILE'] = RepositoryInstance::getOperador()->getOperatorAccessProfile($nrParcNegocio, $organizationId);
        $customAccessProfile        = $this->getAccessProfileByExternalFunction($organizationId, $userId);
        $userData['ACCESS_PROFILE'] = empty($customAccessProfile) ? $userData['ACCESS_PROFILE'] : $customAccessProfile;

        $expiredPassword = $this->validations($userData, $password, $productId, $requesterUrl, $attempts);
        $filter          = $this->getLoginParams($teknisaOrgId, $productId);
        try {
            $loginResponse = Utilities::callWebService(null, '/productLicense', $filter, 'userData', $this->loginClientSecret, $this->loginUrl);
        } catch(\Exception $e) {
            if(in_array($e->getCode(), $this->licenseExceptions)) {
                throw $e;
            }
            $loginResponse = null;
        }
        if(empty($loginResponse)) {
            $productPermission = $this->getProductPermissions($productId, $organizationId);
            $loginResponse = array(
                "TOKEN"             => null,
                "REDIRECT_URL"      => null,
                "VISIBLE_ELEMENTS"  => $productPermission['elements'],
                "EXPIRED_ELEMENTS"  => isset($productPermission['expiredElements']) ? $productPermission['expiredElements'] : array(),
                "REPOSITORIES_INFO" => $productPermission['migrationRepositories']
            );
        } else {
            $relatedOrgIds = isset($loginResponse['RELATED_ORG_IDS']) ? $loginResponse['RELATED_ORG_IDS'] : array();
            $this->saveOAccTekMgr($loginResponse['PRODUCTS_PERMISSION'], $organizationId, $userId, $relatedOrgIds);
            $variableMeasure = $this->getVariableMeasureRegisters($loginResponse['VARIABLE_MEASURE'], $productId, $userId, $userName, $organizationId, $teknisaOrgId);
            if(!empty($variableMeasure)) {
                $this->saveVariableMeasure($variableMeasure, $loginResponse['TOKEN']);
            }
        }

        $loginResponse['TOKEN'] = InstanceProvider::getOAuth()->grantAccessToken(
            $this->libClientId,
            $this->libClientSecret,
            array('sessionId' => InstanceProvider::getSession()->getId()),
            $loginResponse['TOKEN']
        );

        return $this->factorySessionData($userData, $loginResponse, $expiredPassword);
    }

    /**
     * @param $userData
     * @param $loginResponse
     * @param $expiredPassword
     * @return array
     */
    private function factorySessionData($userData, $loginResponse, $expiredPassword) {
        $user = $userData['USER'];
        $userInfo = array(
            "USER_ID"           => $user['ID'],
            "USER_EMAIL"        => $user['EMAIL'],
            "TOKEN"             => $loginResponse['TOKEN'],
            "REDIRECT_URL"      => $loginResponse['REDIRECT_URL'],
            "VISIBLE_ELEMENTS"  => $loginResponse['VISIBLE_ELEMENTS'],
            "EXPIRED_ELEMENTS"  => isset($loginResponse['EXPIRED_ELEMENTS']) ? $loginResponse['EXPIRED_ELEMENTS'] : [],
            "REPOSITORIES_INFO" => $loginResponse['REPOSITORIES_INFO'],
            "USER_DATA"         => $userData
        );
        $encryptedSession = LibUtilities::zeedhiEncrypt(json_encode($userInfo));

        $sessionData = array(
            "USER_ID"           => $user['ID'],
            "USER_NAME"         => $user['NAME'],
            "USER_EMAIL"        => $user['EMAIL'],

            "REDIRECT_URL"      => $loginResponse['REDIRECT_URL'],
            "TOKEN"             => $loginResponse['TOKEN'],
            "USER"              => $encryptedSession,
            "EXPIRED_PASSWORD"  => $expiredPassword,
            "LOGGED"            => true,
        );

        if(!empty($loginResponse['REDIRECT_URL'])) {
            $redirectionToken = $this->startRedirectionProductSession($encryptedSession, $loginResponse['REDIRECT_URL']);
            $sessionData['REDIRECTION_TOKEN'] = $redirectionToken;
        }

        return $sessionData;
    }

    /**
     * @param $productId
     * @param $organizationId
     * @return array
     * @throws \Exception
     */
    public function getProductPermissions($productId, $organizationId)
    {
        $oAccTekMgr = RepositoryInstance::getOAccTekMgr()->getOAccTekMgrByNrOrg($organizationId);
        if(empty($oAccTekMgr)) {
            Exception::productIsNotAuthorized();
        }
        $oAccTekDecrypted = LibUtilities::zeedhiDecrypt($oAccTekMgr['OACCTEK'], LibUtilities::LOGIN_CRYPT_STRING);

        $orgAccess = json_decode($oAccTekDecrypted, true);
        if (empty($orgAccess)) {
            Exception::orgAccessDecryptFail();
        }

        foreach ($orgAccess['products'] as $orgProduct) {
            if (isset($orgProduct['productId']) && $orgProduct['productId'] == $productId) {
                $productPermission = $orgProduct;
                break;
            }
        }

        if(empty($productPermission)) {
            Exception::productIsNotAvailable();
        }
        if(empty($productPermission['elements'])) {
            Exception::menuItemsNotFound();
        }

        return $productPermission;
    }

    /**
     * @param $user
     * @param $password
     * @return array
     */
    public function validateUser($user, $password) {
        $userData = array();
        $userDataProvider = InstanceProvider::getLibUserDataProvider();

        $operator = InstanceProvider::getLibUserDataRepository()->getOperatorByCdoperador($user);
        $userDataProvider->getPasswordParameters($operator['CDOPERADOR'], $operator['NRORG'], $userData);
        $passwordParams = $userData['PASSWORD_PARAMETERS'];

        if(!$userDataProvider->validateAD($passwordParams, $operator['NMACESSO'], $password, $operator['NRORG'])) {
            $this->passwordValidation($operator, $password, $passwordParams, false);
        }

        InstanceProvider::getLibEnvironment()->setSessionExpired(false);
        return array(true);
    }

    /**
     * @param $organizationId
     * @param $productId
     * @return array
     * @throws \Exception
     */
    protected function getLoginParams($organizationId, $productId) {
        $menus = InstanceProvider::getLibMenuService()->buildMenu(true, array(Menu::ALL_ELEMENTS_VISIBLE), null, null, null, null);
        $elementIds = array();
        $this->getElementIds($menus, $elementIds);
        return array (
            "requestType" => "FilterData",
            "filter" => array( array(
                'name' => 'ORGANIZATION_ID', 'operator' => '=', 'value' => $organizationId
            ), array (
                'name' => 'PRODUCT_ID',      'operator' => '=', 'value' => $productId
            ), array (
                'name' => 'ELEMENT_IDS',     'operator' => '=', 'value' => $elementIds
            ) )
        );
    }

    /**
     * @param $menus
     * @param $elementIds
     */
    protected function getElementIds($menus, &$elementIds) {
        $menus = isset($menus['MENUS']) ? $menus['MENUS'] : $menus;
        foreach($menus as $menu) {
            if(isset($menu['id'])) {
                array_push($elementIds, $menu['id']);
            }
            if(isset($menu['menus'])) {
                $this->getElementIds($menu['menus'], $elementIds);
            }
        }
    }

    /**
     * @param $userData
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @return bool|string
     * @throws \Exception
     */
    protected function validations($userData, $password, $productId, $requesterUrl, $attempts) {
        if(empty($userData) || empty($userData['USER'])) {
            Exception::invalidPassword();
        }

        $otherSessionValues = $userData['OTHER_SESSION_VALUES'];
        $passwordStatus = $this->userValidation($userData['USER'], $productId);
        $this->accessProfileValidation($userData['ACCESS_PROFILE'], $otherSessionValues);
        $this->validateIp($userData['IP_RESTRICTION'], LibUtilities::getCurrentIp());
        $this->validateDomain($requesterUrl);

        $expiredPassword = $this->passwordValidation($userData['USER'], $password, $userData['PASSWORD_PARAMETERS'],
            true, $attempts, $otherSessionValues['ACL_ADMINISTRATOR'], $passwordStatus);

        return $expiredPassword;
    }

    /**
     * @param $user
     * @param $productId
     */
    protected function userValidation($user, $productId) {
        if(!isset($user['IS_ACTIVE']) || $user['IS_ACTIVE'] !== self::IS_ACTIVE_POSITIVE) {
            Exception::userInactive();
        }
        if(empty($user['ORGANIZATION_ID'])) {
            Exception::userNotOrganization();
        }
        if (empty($productId)) {
            Exception::productIdNotFound();
        }

        $operatorParams = RepositoryInstance::getOperador()->getOperatorParams($user['ID'], $user['ORGANIZATION_ID']);
        if($operatorParams['IDSTATUSENHA'] == self::INVALID_PASSWORD_STATUS) {
            Exception::accessBlocked();
        }

        return $operatorParams['IDSTATUSENHA'];
    }

    /**
     * @param $accessProfile
     * @param $otherSessionValues
     */
    protected function accessProfileValidation($accessProfile, $otherSessionValues) {
        if(empty($accessProfile) &&
            ((isset($otherSessionValues['IDUTCTRLACESWEB']) && $otherSessionValues['IDUTCTRLACESWEB'] == 'S') ||
                (isset($otherSessionValues['IDUTCTRLACES'])    && $otherSessionValues['IDUTCTRLACES']    == 'S'))
        ) {
            Exception::userNotAccessProfile();
        }
    }

    /**
     * @param $user
     * @param $password
     * @param $passwordParams
     * @param bool $isValidateExpired
     * @param $attempts
     * @param $isAdmin
     * @return bool|string
     */
    protected function passwordValidation($user, $password, $passwordParams, $isValidateExpired = true, $attempts = null, $isAdmin = null, $passwordStatus = null) {
        $isAD = $passwordParams && $passwordParams['USES_ACTIVE_DIRECTORY'] == self::USES_ACTIVE_DIRECTORY_POSITIVE;

        /** If exists attempts, this method called in login.
         *  Else, called by expired session */
        if(empty($attempts)) {
            $paramSenha = null;
        } else {
            $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($user['ORGANIZATION_ID']);
            $paramSenha    = RepositoryInstance::getParamSenha()->getParamSenhaByNrOrg($user['ORGANIZATION_ID'], $entityManager);
        }

        if(!$user['PASSWORD_MD5']) {
            $this->validateDelphiPassword($user, $password, $isAD, $attempts, $paramSenha, $isAdmin);
        } else {
            $this->validateWebPassword($user, $password, $isAD, $attempts, $paramSenha, $isAdmin);
        }

        if($passwordStatus && $passwordStatus == self::BLOCKED_PASSWORD_STATUS) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if($isValidateExpired && $passwordParams) {
            return $this->validateExpiredPassword($passwordParams, $paramSenha);
        }

        return false;
    }

    /**
     * @param $user
     * @param $loginPassword
     * @param $isAD
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    protected function validateDelphiPassword($user, $loginPassword, $isAD, $attempts, $paramSenha, $isAdmin) {
        $loginPasswordEncrypted = !$isAD ? LibUtilities::getDelphiPasswordEncrypted($loginPassword) : $loginPassword;
        if($user['PASSWORD_CRYPTO_PLUS'] != $loginPasswordEncrypted) {
            $this->throwInvalidPasswordError($user['ID'], $user['ORGANIZATION_ID'], $attempts, $paramSenha, $isAdmin);
        }
    }

    /**
     * @param $user
     * @param $loginPassword
     * @param $isAD
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    protected function validateWebPassword($user, $loginPassword, $isAD, $attempts, $paramSenha, $isAdmin) {
        $loginPasswordEncrypted = !$isAD ? LibUtilities::encrypt($loginPassword) : $loginPassword;
        if($user['PASSWORD_MD5'] != $loginPasswordEncrypted) {
            $this->throwInvalidPasswordError($user['ID'], $user['ORGANIZATION_ID'], $attempts, $paramSenha, $isAdmin);
        }
    }

    /**
     * @param $userId
     * @param $organizationId
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    public function throwInvalidPasswordError($userId, $organizationId, $attempts, $paramSenha, $isAdmin) {
        if(empty($paramSenha) || $isAdmin) {
            Exception::invalidPassword();
        } else {
            $nrErrors = $paramSenha[0]['NRERROSENHA'];
            if(!empty($nrErrors) && $attempts == $nrErrors - 1) {
                Exception::invalidPasswordBlockUser();
            } else if(!empty($nrErrors) && $attempts >= $nrErrors) {
                RepositoryInstance::getOperador()->saveOperatorPasswordStatus($userId, $organizationId, self::INVALID_PASSWORD_STATUS);
                Exception::accessBlockedByAttempts();
            } else {
                Exception::invalidPasswordCode32();
            }
        }
    }

    /**
     * @param $passwordParams
     * @param $paramSenha
     * @return bool|string
     */
    protected function validateExpiredPassword($passwordParams, $paramSenha) {
        if($passwordParams['IDENTIFICATION'] == self::ID_RESET_PASSWORD) {
            return self::RESET_PASSWORD_IDENTIFICATION;
        }
        if(empty($passwordParams['UPDATE_DATE'])) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if(!empty($passwordParams['PASSWORD_EXPIRATION_DAYS']) && $passwordParams['USES_PASSWORD_CONTROL'] == self::USES_PASSWORD_CONTROL_POSITIVE) {
            $expireDate = \DateTime::createFromFormat('d/m/Y', $passwordParams['UPDATE_DATE']);
            $interval = 'P' . ($passwordParams['PASSWORD_EXPIRATION_DAYS']) . self::DATE_TYPE_DAY;
            $expireDate->add(new \DateInterval($interval));
            $actualDate = new \DateTime();
            $intervalExp  = Utilities::getDateIntervalInSeconds($expireDate, $actualDate);
            $intervalDays = Utilities::getDateIntervalInSeconds($actualDate, $expireDate);
            if($intervalDays > 0) {
                $intervalDays = $intervalDays / 86400;
            } else if($intervalDays < 0) {
                $intervalDays = null;
            }
            if ($intervalExp > 0) {
                return self::EXPIRED_PASSWORD_IDENTIFICATION;
            } else if($intervalDays == 0) {
                return self::PASSWORD_EXPIRE_TODAY_IDENTIFICATION;
            } else if($intervalDays && $intervalDays <= $paramSenha[0]['NRDIASAVISO']) {
                return self::PASSWORD_WILL_EXPIRE_IDENTIFICATION . '_' . $intervalDays;
            }
        }
        return false;
    }

    /**
     * @param $ipRestriction
     * @param $ip
     * @throws \Exception
     */
    protected function validateIp($ipRestriction, $ip) {
        if(!empty($ipRestriction) && $ipRestriction['USES_IP_RESTRICTION'] == self::USES_IP_RESTRICTION_POSITIVE) {
            $isValidIp = false;
            if(Utilities::isIpv4($ip)) {
                foreach($ipRestriction['IPV4'] as $ipV4) {
                    if(Utilities::validateIpInRange($ip, $ipV4)) {
                        $isValidIp = true;
                        break;
                    }
                }
            } else {
                foreach($ipRestriction['IPV6'] as $ipV6) {
                    if($ip == $ipV6) {
                        $isValidIp = true;
                        break;
                    }
                }
            }
            if(!$isValidIp) {
                Exception::ipIsOutOfRange($ip);
            }
        }
    }

    /**
     * @param $url
     * @return bool
     */
    protected function validateDomain($url) {
        if(Utilities::getUseDomainFilterParameter()) {
            $validDomains = explode(';', Utilities::getDomainFilterParameter());
            foreach($validDomains as $domain) {
                if(!empty($domain) && strpos($url, $domain) !== false) {
                    return true;
                }
            };
            Exception::invalidDomain();
        }
        return true;
    }

    /**
     * @param array $productsPermission
     * @param $organizationId
     * @param $userId
     * @param $relatedOrgIds
     * @throws \Exception
     */
    public function saveOAccTekMgr($productsPermission, $organizationId, $userId, $relatedOrgIds) {
        $oAccTekMgrRepository = RepositoryInstance::getOAccTekMgr();
        $licences   = array('products' => $productsPermission, 'macAddress' => Utilities::getMacAddress(), 'relatedOrgIds' => $relatedOrgIds);
        $licences   = LibUtilities::zeedhiEncrypt(json_encode($licences), LibUtilities::LOGIN_CRYPT_STRING);
        $oAccTekMgr = $oAccTekMgrRepository->getOAccTekMgrByNrOrg($organizationId);

        if(empty($oAccTekMgr)) {
            $oAccTekMgrRepository->insertOAccTekMgr($organizationId, $userId, self::RECURRENCE_DAYS, $licences);
        } else {
            $lastupdate = \DateTime::createFromFormat('d/m/Y', $oAccTekMgr['DTULTATU']);
            $interval   = 'P' . $oAccTekMgr['RECURRENCE_DAYS'] . self::DATE_TYPE_DAY;
            $lastupdate->add(new \DateInterval($interval));
            $actualDate = new \DateTime();

            $oAccTek = @json_decode(LibUtilities::zeedhiDecrypt($oAccTekMgr['OACCTEK'], LibUtilities::LOGIN_CRYPT_STRING), true);
            $actualRelatedOrgIds = isset($oAccTek['relatedOrgIds']) ? $oAccTek['relatedOrgIds'] : array();
            $isDiffOrgIds = !empty(array_diff($relatedOrgIds, $actualRelatedOrgIds)) || !empty(array_diff($actualRelatedOrgIds, $relatedOrgIds));

            if (Utilities::getDateIntervalInSeconds($lastupdate, $actualDate) > 0 || $isDiffOrgIds) {
                $oAccTekMgrRepository->updateOAccTekMgr($organizationId, $userId, self::RECURRENCE_DAYS, $licences);
            }
        }
    }

    /**
     * @param $variableMeasure
     * @param $product
     * @param $userId
     * @param $userName
     * @param $organizationId
     * @param $teknisaOrgId
     * @return array
     * @throws \Exception
     */
    protected function getVariableMeasureRegisters($variableMeasure, $product, $userId, $userName, $organizationId, $teknisaOrgId) {
        $variableMeasureToUpdate = array();
        foreach($variableMeasure as $variableMeasureRegister) {
            $count = $log = $error = null;
            $countOutput = $this->execPHPScriptVariableMeasure($variableMeasureRegister, $organizationId, false);
            $countOutputDecoded = @json_decode($countOutput, true);
            if(empty($countOutputDecoded)) {
                $error = $countOutput;
            } else {
                $count = isset($countOutputDecoded[0]) && is_array($countOutputDecoded[0]) ?
                    $countOutputDecoded[0]['COUNT'] : $countOutputDecoded['COUNT'];
            }
            if(!empty($error)) {
                $variableMeasureToUpdate[] = $this->buildVariableMeasureRow($variableMeasureRegister, $count, $log, $product,
                    $userId, $userName, $teknisaOrgId, $error);
            } else if(empty($variableMeasureRegister['ACTUAL_COUNT']) || $variableMeasureRegister['ACTUAL_COUNT'] < $count) {
                $logOutput = $this->execPHPScriptVariableMeasure($variableMeasureRegister, $organizationId, true);
                $logOutputDecoded = @json_decode($logOutput, true);
                if(empty($logOutputDecoded)) {
                    $error = $logOutput;
                } else {
                    $log = $logOutputDecoded;
                }
                $variableMeasureToUpdate[] = $this->buildVariableMeasureRow($variableMeasureRegister, $count, $log, $product,
                    $userId, $userName, $teknisaOrgId, $error);
            }
        }
        return $variableMeasureToUpdate;
    }

    protected function saveVariableMeasure($variableMeasure, $token) {
        $params = array(
            "requestType" => "DataSet",
            "dataset"     => $variableMeasure
        );
        return Utilities::callWebService($token, '/variableMeasure/save', $params, 'variableMeasureResult',
            $this->loginClientSecret, $this->loginUrl);
    }

    /**
     * @param array $variableMeasureRegister
     * @param $count
     * @param $log
     * @param $product
     * @param $userId
     * @param $userName
     * @param $organizationId
     * @param $error
     * @return array
     */
    protected function buildVariableMeasureRow($variableMeasureRegister, $count, $log, $product, $userId, $userName, $organizationId, $error) {
        return array(
            'COUNT' => $count,
            'ERROR' => $error,
            'LOG'   => $log,

            'PRODUCT_ID'   	  => $product,
            'OPERATOR_CODE'   => $userId,
            'OPERATOR_NAME'   => $userName,
            'ORGANIZATION_ID' => $organizationId,

            'ID'              => $variableMeasureRegister['ID'],
            'PRODUCT_NAME' 	  => $variableMeasureRegister['NMPRODUTO'],
            'MEASURE_CODE' 	  => $variableMeasureRegister['CDTPMEDIDASERV'],
            'MEASURE_NAME' 	  => $variableMeasureRegister['DSTPMEDIDASERV'],
            'INIT_COMPETENCE' => $variableMeasureRegister['DTINICOMPET'],
            'END_COMPETENCE'  => $variableMeasureRegister['DTFINCOMPET'],

            'CONTRACT_REFERENCE' => $variableMeasureRegister['DSREFERENCIAORCA'],
            'CONTRACT_ID'        => $variableMeasureRegister['NRSEQCONTRATO'],
            'CONTRACT_EXP_ID'    => $variableMeasureRegister['NRVENCPREVCONT'],
        );
    }

    /**
     * @param $variableMeasureRegister
     * @param $organizationId
     * @param $isLogScript
     * @return string
     * @throws \Exception
     */
    protected function execPHPScriptVariableMeasure($variableMeasureRegister, $organizationId, $isLogScript) {
        $scriptContents   = $isLogScript ? $variableMeasureRegister['PHP_SCRIPT_LOG'] : $variableMeasureRegister['PHP_SCRIPT_COUNT'];
        $contractInitDate = $variableMeasureRegister['DTINICOMPET'];
        $contractEndDate  = $variableMeasureRegister['DTFINCOMPET'];
        $contractId       = $variableMeasureRegister['NRSEQCONTRATO'];
        $newEntityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($organizationId);
        if($newEntityManager) {
            $connectionParams = $newEntityManager->getConnection()->getParams();
        } else {
            $connectionParams = InstanceProvider::getEntityManager()->getConnection()->getParams();
        }
        $connectionParams = json_encode($connectionParams);
        $fileName = $isLogScript ? "/phpFileLog.php" : "/phpFileCount.php";
        $fileName = Utilities::getTmpPathParameter() . $fileName;
        file_put_contents($fileName, $scriptContents);
        $params = " '$connectionParams' '$organizationId' '$contractInitDate' '$contractEndDate' '$contractId' ";
        $output = trim(shell_exec("php " . $fileName . $params . "2>&1"));
        unlink($fileName);
        return $output;
    }

    public function getProductLanguages($productId) {
        $filter = array(
            "requestType" => "FilterData",
            "filter" => array( array(
                'name' => 'PRODUCT_ID', 'operator' => '=', 'value' => $productId
            ) )
        );
        return Utilities::callWebService(null, '/productLanguages', $filter, 'languages', $this->loginClientSecret, $this->loginUrl);
    }

    /**
     * @return array
     */
    protected function readOAccCustom() {
        $oAccCustom = array();
        try {
            $fileDir = Utilities::getOAccCustomParameter();
            if (file_exists($fileDir)) {
                $fileContents  = file_get_contents($fileDir);
                $fileDecrypted = LibUtilities::zeedhiDecrypt($fileContents, LibUtilities::LOGIN_CRYPT_STRING);
                $fileDecoded   = @json_decode($fileDecrypted, true);
                if(isset($fileDecoded['organizationId']) &&
                    isset($fileDecoded['macAddress']) &&
                    is_numeric($fileDecoded['organizationId'])) {
                    $oAccCustom = $fileDecoded;
                }
            }
        } catch(\Exception $e) {}

        if(!empty($oAccCustom)) {
            $macs = explode(';', $oAccCustom['macAddress']);
            if(!in_array(Utilities::getMacAddress(), $macs)) {
                Exception::incorrectMacAddressOAccCustom();
            }
        }
        return $oAccCustom;
    }

    /**
     * @param $organizationId
     * @param $macAddress
     */
    public function genOAccCustom($organizationId, $macAddress) {
        if(!empty($organizationId) && !empty($macAddress)) {
            $oAccCustom = array(
                'organizationId' => $organizationId,
                'macAddress' => $macAddress
            );
            $oAccCrypted = LibUtilities::zeedhiEncrypt(json_encode($oAccCustom), LibUtilities::LOGIN_CRYPT_STRING);
            file_put_contents(Utilities::getOAccCustomParameter(), $oAccCrypted);
            echo "File created!\n";
        } else {
            if(empty($organizationId)) echo "Organization ID not found.\n";
            if(empty($macAddress))     echo "MAC Address not found.\n";
        }
    }

    private function startRedirectionProductSession($user, $redirectUrl) {
        $url    = $redirectUrl . '/backend/index.php';
        $token  = Utilities::getLibToken($url);
        $filter = $this->getRedirectionProductRequestParams($token, $user);
        Utilities::callWebService($token, '/lib_startSession', $filter, 'system', $this->libClientSecret, $url);
        return $token;
    }

    public function getRedirectionProductRequestParams($token, $user) {
        return array(
            "requestType" => "FilterData",
            "filter" => array( array(
                'name' => 'USER',        'operator' => '=', 'value' => $user
            ), array(
                'name' => 'IS_LOGGED',   'operator' => '=', 'value' => true
            ), array(
                'name' => 'LOGIN_TOKEN', 'operator' => '=', 'value' => $token
            ) )
        );
    }

    private function getAccessProfileByExternalFunction($nrOrg, $operatorCode) {
        $function = Utilities::getAccessProfileFunctionParameter();
        if(empty($function)) {
            return null;
        } else {
            return InstanceProvider::getEntityManager()->getRepository(Utilities::GENERAL_ENTITY_NAMESPACE . 'Operador')
                ->$function($nrOrg, $operatorCode);
        }
    }

    public function testLicenseConnection() {
        $error = false;
        try {
            Utilities::callWebService(null, '/testConnection', array(), 'testConnection',
                $this->loginClientSecret, $this->loginUrl, false, true);
        } catch(\Exception $e) {
            $error = $e->getMessage();
        }
        return array (
            'ERROR' => $error,
            'URL'   => explode('/backend', $this->loginUrl)[0]
        );
    }

    private function getTeknisaOrganizationId($userData) {
        $oAccCustom     = $this->readOAccCustom();
        $nrOrgOper      = $userData['USER']['NRORGOPER'];
        $organizationId = $userData['USER']['ORGANIZATION_ID'];

        if(!empty($oAccCustom)) {
            return $oAccCustom['organizationId'];
        }
        if(!empty($nrOrgOper)) {
            return $nrOrgOper;
        }
        return $organizationId;
    }

    private function getOrganizationId($userData) {
        $nrOrgOper      = $userData['USER']['NRORGOPER'];
        $organizationId = $userData['USER']['ORGANIZATION_ID'];

        if(!empty($nrOrgOper)) {
            return $nrOrgOper;
        }
        return $organizationId;
    }
}